
//
// wthread.h: This is a set of classes and functions that
//            wrap a Win32 thread. It allows you to start
//            a thread on any member function of any object
//            provided it has a certain signature:
//              DWORD ObjT::Func ( ArgT );
//            
//            It has three advantages:
//            - It allows you to create a thread on any object 
//              instance, on any function with the signature:
//              DWORD func(ArgT arg);
//            - it's type safe.
//            - it has pluggable Thread creators
//
// This file is part of the Winterdom library
// Copyright (C) 2000-2001, Tomas Restrepo (tomasr@mvps.org)
//            
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#ifndef WTHREAD_H__INCLUDED
#define WTHREAD_H__INCLUDED

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "wctypes.h"
#include "wexcept.h"

namespace Winterdom { 
   namespace Runtime {
      namespace Threading {

   using namespace Winterdom;

   //
   // CrtThreader is a simple class
   // to create a thread using the CRT's
   // _beginthreadex() call
   //
   struct CrtThreader
   {
      typedef unsigned (__stdcall *PTHREADPROC)(void*);

      static HANDLE CreateThread ( 
            LPSECURITY_ATTRIBUTES  attributes,
            DWORD                  stackSize,
            LPTHREAD_START_ROUTINE threadProc,
            LPVOID                 param,
            DWORD                  creationFlags,
            LPDWORD                id
         )
      {
         return (HANDLE)_beginthreadex (
                           attributes, stackSize,
                           (PTHREADPROC)threadProc,
                           (void*)param, creationFlags,
                           (UINT*)id
                        );
      }
   };

   //
   // Win32Threader is a simple class
   // to create a thread using the Win32
   // api CreateThread()
   //
   struct Win32Threader
   {
      static HANDLE CreateThread ( 
            LPSECURITY_ATTRIBUTES  attributes,
            DWORD                  stackSize,
            LPTHREAD_START_ROUTINE threadProc,
            LPVOID                 param,
            DWORD                  creationFlags,
            LPDWORD                id
         )
      {
         return ::CreateThread (
                     attributes, stackSize,
                     threadProc, (void*)param, 
                     creationFlags, id
                  );
      }
   };


   //
   // Thread is a simple class that wraps a 
   // thread handle and ID. It wraps the 
   // most commonly used thread apis, like
   // SuspendThread() and ResumeThread().
   // Plus, it has value semantics and full
   // copy constructor and assignment operator.
   //
   class Thread
   {
   private: 
      HANDLE   m_hThread;
      DWORD    m_id;
   public:
      Thread ( HANDLE h = NULL, DWORD id = 0)
         : m_hThread(h), m_id(id)
      {
      }
      ~Thread() 
      {
         if ( m_hThread != NULL )
            CloseHandle ( m_hThread );
      }

      // copy contructor and assignment
      Thread ( const Thread& t )
      {
         if ( !CopyHandle ( t.Handle() ) )
            throw Support::Win32Exception(GetLastError());
         m_id = t.Id();
      }
      Thread& operator= ( const Thread& t )
      {
         if ( this != &t )
         {
            if ( !CopyHandle ( t.Handle() ) )
               throw Support::Win32Exception(GetLastError());
            m_id = t.Id();
         }
         return *this;
      }

      // == accessors ==
      HANDLE Handle() const
      {
         return m_hThread;
      }
      void SetHandle ( HANDLE h )
      {
         m_hThread = h;
      }
      DWORD Id() const
      {
         return m_id;
      }
      void SetId ( DWORD id )
      {
         m_id = id;
      }
      bool IsValid() const
      {
         return (m_hThread != NULL);
      }

      // == thread api wrappers ==
      DWORD Resume()
      {
         return ResumeThread ( m_hThread );
      }
      DWORD Suspend()
      {
         return SuspendThread ( m_hThread );
      }
      DWORD WaitOn ( DWORD timeout = INFINITE )
      {
         return WaitForSingleObject ( m_hThread, timeout );
      }

   private:
      bool CopyHandle ( HANDLE h )
      {
         BOOL b = ::DuplicateHandle ( 
            ::GetCurrentProcess(), h,
            ::GetCurrentProcess(), &m_hThread,
            0, FALSE, DUPLICATE_SAME_ACCESS
            );
         return (b != FALSE);    
      }
   };

   // 
   // class ThreadImpl wraps a Win32 Thread object.
   // You use it by providing the function you want
   // called in the thread on the CreateThread call
   //
   // It has three advantages:
   // - It allows you to create a thread on any object 
   //   instance, on any function with the signature:
   //   DWORD func(ArgT arg);
   // - it's type safe.
   // - it has pluggable Thread creators
   //
   template <
      typename ObjT,       // Object Type
      typename ArgT,       // Thread argument type
      typename CreatorT    // Thread creator
   > class ThreadImpl
   {
   public:
      typedef DWORD (ObjT::*pfunc)(ArgT);

      ThreadImpl() { }

   public:
      Thread CreateThread ( ObjT& obj, pfunc func, 
                            ArgT arg, bool CreateSuspended=false )
      {
         HANDLE hThread; DWORD id;

         ThreadParams* pt = new ThreadParams ( &obj, func, arg );
         if ( pt == 0 )
            throw Support::GenericException(_T("Memory Allocation Failed"));

         hThread = CreatorT::CreateThread (
                        NULL, 0, ThreadProc,
                        (void*)pt,
                        CreateSuspended ? CREATE_SUSPENDED : 0,
                        &id
                     );
         if ( hThread == NULL )
         {
            delete pt;
            throw Support::Win32Exception(GetLastError());
         }

         return Thread(hThread, id);
      } // Create()

   private:
      // hide copy contructor and assignment
      ThreadImpl ( const ThreadImpl& t );
      ThreadImpl& operator= ( const ThreadImpl& t );


      struct ThreadParams {
         ObjT*  self;
         pfunc  func;
         ArgT   arg;
         ThreadParams(ObjT* s, pfunc f, ArgT a)
            : self(s), func(f), arg(a) {}
      };

      // internal thread procedure
      static DWORD __stdcall ThreadProc ( void* param )
      {
         ThreadParams* pt = static_cast<ThreadParams*>(param);
         DWORD ret = ((pt->self)->*(pt->func))(pt->arg);
         delete pt;
         return ret;
      }

   };

   // 
   // GlobalAdaptor is simple object used
   // to adapt ThreadImpl so that we can also
   // start threads on stand-alone global functions
   // notice how we template it so that it can
   // accomodate both __cdecl and __sdtcall
   // funtions
   //
   template <typename FuncT, typename ArgT>
   class GlobalAdaptor 
   {
   public:
      explicit GlobalAdaptor ( FuncT func ) : m_func(func) {};
      DWORD Proc ( ArgT arg ) {
         DWORD ret = m_func(arg);
         delete this;
         return ret;
      }
   private:
      FuncT m_func;
   };

   //
   // our thread creation functions : they allow
   // us to have automatic type discovery!
   //

   // on objects
   template <typename CreatorT, class ObjT, class FuncT, class ArgT>
   Thread RunOnThread ( ObjT& obj, FuncT pfunc, ArgT arg,  
                        bool CreateSuspended = false )
   {
      ThreadImpl<ObjT, ArgT, CreatorT> t;
      return t.CreateThread ( obj, pfunc, arg, CreateSuspended );
   }

   // stand alone
   template <typename CreatorT, class FuncT, class ArgT>
   Thread RunOnThread ( FuncT pfunc, ArgT arg, 
                        bool CreateSuspended = false )
   {
      typedef GlobalAdaptor<FuncT, ArgT> AdaptorType;
      AdaptorType* adaptor = new AdaptorType(pfunc);

      ThreadImpl<AdaptorType, ArgT, CreatorT> t;
      return t.CreateThread ( 
                  *adaptor, &AdaptorType::Proc, 
                  arg, CreateSuspended 
               );
   }


      } // namespace Threading
   } // namespace Runtime
} // namespace Winterdom

#endif // WTHREAD_H__INCLUDED
